from dataclasses import dataclass
from pathlib import Path
from typing import Optional, List, Dict, Tuple

import nassl

from sslyze.errors import TlsHandshakeFailed
from sslyze.plugins.certificate_info._cert_chain_analyzer import (
    CertificateDeploymentAnalyzer,
    CertificateDeploymentAnalysisResult,
)
from sslyze.plugins.certificate_info._cli_connector import _CertificateInfoCliConnector
from sslyze.plugins.certificate_info._get_cert_chain import get_certificate_chain, ArgumentsToGetCertificateChain
from sslyze.plugins.certificate_info.trust_stores.trust_store import TrustStore
from sslyze.plugins.certificate_info.trust_stores.trust_store_repository import TrustStoresRepository
from sslyze.plugins.plugin_base import (
    ScanCommandImplementation,
    ScanJob,
    ScanCommandResult,
    ScanCommandExtraArgument,
    ScanJobResult,
)
from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum


@dataclass(frozen=True)
class CertificateInfoExtraArgument(ScanCommandExtraArgument):
    """Additional configuration for running the certificate_info scan command.

    Attributes:
        custom_ca_file: The path to a custom trust store file to use for certificate validation. The file should contain
            PEM-formatted root certificates.
    """

    custom_ca_file: Path

    def __post_init__(self) -> None:
        if not self.custom_ca_file.is_file():
            raise ValueError(f'Could not open supplied CA file at "{self.custom_ca_file}"')


@dataclass(frozen=True)
class CertificateInfoScanResult(ScanCommandResult):
    """The result of retrieving and analyzing a server's certificates to verify their validity.

    Attributes:
        hostname_used_for_server_name_indication: The hostname sent by SSLyze as the Server Name Indication extension.
        certificate_deployments: A list of leaf certificates detected by SSLyze and the corresponding analysis. Most
            servers only deploy one leaf certificate, but some websites (such as Facebook) return different leaf
            certificates depending on the client, as a way to maximize compatibility with older clients/devices.
    """

    hostname_used_for_server_name_indication: str
    certificate_deployments: List[CertificateDeploymentAnalysisResult]


class CertificateInfoImplementation(ScanCommandImplementation[CertificateInfoScanResult, None]):
    """Retrieve and analyze a server's certificate(s) to verify its validity.
    """

    cli_connector_cls = _CertificateInfoCliConnector

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[CertificateInfoExtraArgument] = None
    ) -> List[ScanJob]:
        custom_ca_file = extra_arguments.custom_ca_file if extra_arguments else None

        # Try to retrieve different certificates from the server by having SSLyze's TLS handshake look like different
        # kinds of clients
        call_arguments: List[ArgumentsToGetCertificateChain] = []
        if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
            # Get the default certificate chain sent to clients using TLS 1.3
            call_arguments.append((server_info, custom_ca_file, TlsVersionEnum.TLS_1_3, None))

            # Get the other certificate chains sent to clients using TLS 1.2 that support or don't support RSA
            call_arguments.append((server_info, custom_ca_file, TlsVersionEnum.TLS_1_2, "RSA"))
            call_arguments.append((server_info, custom_ca_file, TlsVersionEnum.TLS_1_2, "ALL:-RSA"))
        else:
            # Get the certificate chains sent to clients that support or don't support RSA
            call_arguments.append((server_info, custom_ca_file, None, None))
            call_arguments.append((server_info, custom_ca_file, None, "RSA"))
            call_arguments.append((server_info, custom_ca_file, None, "ALL:-RSA"))

        # The custom_ca_file is not needed by get_certificate_chain() but we have to pass it anyway so we can eventually
        # use it in result_for_completed_scan_jobs()
        scan_jobs = [
            ScanJob(function_to_call=get_certificate_chain, function_arguments=call_arg) for call_arg in call_arguments
        ]
        return scan_jobs

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> CertificateInfoScanResult:
        if len(scan_job_results) != 3:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        # Only keep certificate deployments that are different
        # Leaf certificate => certificate chain, OCSP response
        all_configured_certificate_chains: Dict[str, Tuple[List[str], Optional[nassl._nassl.OCSP_RESPONSE]]] = {}
        custom_ca_file = None
        for completed_job in scan_job_results:
            try:
                received_chain_as_pem, ocsp_response, custom_ca_file = completed_job.get_result()
            except TlsHandshakeFailed:
                # Can happen when trying to connect with specific cipher suites (such as RSA or non-RSA)
                continue

            if not received_chain_as_pem:
                raise ValueError("Should never happen")

            all_configured_certificate_chains[received_chain_as_pem[0]] = received_chain_as_pem, ocsp_response

        if not all_configured_certificate_chains:
            raise ValueError("Should never happen")

        # Then validate each certificate/chain deployment
        all_trust_stores = TrustStoresRepository.get_default().get_all_stores()
        if custom_ca_file:
            all_trust_stores.append(TrustStore(custom_ca_file, "Supplied CA file", "N/A"))

        analyzed_deployments = []
        name_to_use_for_hostname_validation = server_info.network_configuration.tls_server_name_indication
        for received_chain_as_pem, ocsp_response in all_configured_certificate_chains.values():
            deployment_analyzer = CertificateDeploymentAnalyzer(
                server_hostname=name_to_use_for_hostname_validation,
                server_certificate_chain_as_pem=received_chain_as_pem,
                server_ocsp_response=ocsp_response,
                trust_stores_for_validation=all_trust_stores,
            )
            analysis_result = deployment_analyzer.perform()
            analyzed_deployments.append(analysis_result)

        # All done
        return CertificateInfoScanResult(
            hostname_used_for_server_name_indication=name_to_use_for_hostname_validation,
            certificate_deployments=analyzed_deployments,
        )



from pathlib import Path
from typing import List, Optional, Tuple

import nassl
from nassl.ssl_client import ClientCertificateRequested

from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum

ArgumentsToGetCertificateChain = Tuple[ServerConnectivityInfo, Optional[Path], Optional[TlsVersionEnum], Optional[str]]


def get_certificate_chain(
    server_info: ServerConnectivityInfo,
    custom_ca_file: Optional[Path],
    tls_version: Optional[TlsVersionEnum],
    openssl_cipher_string: Optional[str],
) -> Tuple[List[str], Optional[nassl._nassl.OCSP_RESPONSE], Optional[Path]]:
    ssl_connection = server_info.get_preconfigured_tls_connection(override_tls_version=tls_version)
    if openssl_cipher_string:
        ssl_connection.ssl_client.set_cipher_list(openssl_cipher_string)

    # Enable OCSP stapling
    ssl_connection.ssl_client.set_tlsext_status_ocsp()

    try:
        ssl_connection.connect()
        ocsp_response = ssl_connection.ssl_client.get_tlsext_status_ocsp_resp()
        received_chain_as_pem = ssl_connection.ssl_client.get_received_chain()

    except ClientCertificateRequested:
        ocsp_response = ssl_connection.ssl_client.get_tlsext_status_ocsp_resp()
        received_chain_as_pem = ssl_connection.ssl_client.get_received_chain()

    finally:
        ssl_connection.close()

    return received_chain_as_pem, ocsp_response, custom_ca_file



from nassl.ssl_client import OpenSslVerifyEnum
from nassl.ssl_client import OpenSslVersionEnum
from nassl.legacy_ssl_client import LegacySslClient


class WorkaroundForTls12ForCipherSuites:
    """Helper to figure out which version of OpenSSL to use for a given TLS 1.2 cipher suite.

    The nassl module supports using either a legacy or a modern version of OpenSSL. When using TLS 1.2, specific cipher
    suites are only supported by one of the two implementation.
    """

    @classmethod
    def requires_legacy_openssl(cls, openssl_cipher_name: str) -> bool:
        # Get the list of all ciphers supported by the legacy OpenSSL
        legacy_client = LegacySslClient(ssl_version=OpenSslVersionEnum.TLSV1_2, ssl_verify=OpenSslVerifyEnum.NONE)
        legacy_client.set_cipher_list("ALL:COMPLEMENTOFALL")
        legacy_ciphers = legacy_client.get_cipher_list()

        # Always use the legacy client if it supports the cipher suite, as the modern OpenSSL (1.1.x) does not support
        # weak ciphers, even with the right compilation options; the handshake fails with a "no ciphers available" error
        # but it actually means that OpenSSL does not support the cipher
        return openssl_cipher_name in legacy_ciphers



from datetime import datetime
from pathlib import Path
from typing import List, Optional
from uuid import UUID

import pydantic

from sslyze import (
    ServerNetworkConfiguration,
    HttpProxySettings,
    ProtocolWithOpportunisticTlsEnum,
    ServerScanStatusEnum,
    ServerConnectivityStatusEnum,
    ClientAuthRequirementEnum,
    ClientAuthenticationCredentials,
)
from sslyze.__version__ import __url__, __version__
from sslyze.plugins.certificate_info.json_output import (
    CertificateInfoExtraArgumentAsJson,
    CertificateInfoScanAttemptAsJson,
)
from sslyze.plugins.compression_plugin import CompressionScanAttemptAsJson
from sslyze.plugins.early_data_plugin import EarlyDataScanAttemptAsJson
from sslyze.plugins.elliptic_curves_plugin import SupportedEllipticCurvesScanAttemptAsJson
from sslyze.plugins.fallback_scsv_plugin import FallbackScsvScanAttemptAsJson
from sslyze.plugins.heartbleed_plugin import HeartbleedScanAttemptAsJson
from sslyze.plugins.http_headers_plugin import HttpHeadersScanAttemptAsJson
from sslyze.plugins.openssl_ccs_injection_plugin import OpenSslCcsInjectionScanAttemptAsJson
from sslyze.plugins.openssl_cipher_suites.json_output import CipherSuitesScanAttemptAsJson
from sslyze.plugins.robot.implementation import RobotScanAttemptAsJson
from sslyze.plugins.session_renegotiation_plugin import SessionRenegotiationScanAttemptAsJson

from sslyze.plugins.session_resumption.json_output import (
    SessionResumptionSupportExtraArgumentAsJson,
    SessionResumptionSupportScanAttemptAsJson,
)
from sslyze import (
    ServerScanResult,
    ServerTlsProbingResult,
)
from sslyze.scanner.models import AllScanCommandsAttempts
from sslyze.server_setting import ConnectionTypeEnum, ServerNetworkLocation


class _BaseModelWithOrmModeAndForbid(pydantic.BaseModel):
    class Config:
        orm_mode = True
        extra = "forbid"  # Fields must match between the JSON representation and the actual objects


class ScanCommandsExtraArgumentsAsJson(_BaseModelWithOrmModeAndForbid):
    # Field is present if extra arguments were provided for the corresponding scan command
    certificate_info: Optional[CertificateInfoExtraArgumentAsJson] = None
    session_resumption: Optional[SessionResumptionSupportExtraArgumentAsJson] = None


class AllScanCommandsAttemptsAsJson(_BaseModelWithOrmModeAndForbid):
    certificate_info: CertificateInfoScanAttemptAsJson
    ssl_2_0_cipher_suites: CipherSuitesScanAttemptAsJson
    ssl_3_0_cipher_suites: CipherSuitesScanAttemptAsJson
    tls_1_0_cipher_suites: CipherSuitesScanAttemptAsJson
    tls_1_1_cipher_suites: CipherSuitesScanAttemptAsJson
    tls_1_2_cipher_suites: CipherSuitesScanAttemptAsJson
    tls_1_3_cipher_suites: CipherSuitesScanAttemptAsJson
    tls_compression: CompressionScanAttemptAsJson
    tls_1_3_early_data: EarlyDataScanAttemptAsJson
    openssl_ccs_injection: OpenSslCcsInjectionScanAttemptAsJson
    tls_fallback_scsv: FallbackScsvScanAttemptAsJson
    heartbleed: HeartbleedScanAttemptAsJson
    robot: RobotScanAttemptAsJson
    session_renegotiation: SessionRenegotiationScanAttemptAsJson
    session_resumption: SessionResumptionSupportScanAttemptAsJson
    elliptic_curves: SupportedEllipticCurvesScanAttemptAsJson
    http_headers: HttpHeadersScanAttemptAsJson

    @classmethod
    def from_orm(cls, all_scan_commands_attempts: AllScanCommandsAttempts) -> "AllScanCommandsAttemptsAsJson":
        all_scan_commands_attempts_json = {}
        for field_name, field in cls.__fields__.items():
            scan_command_attempt = getattr(all_scan_commands_attempts, field_name)

            # Convert the error trace to a string; this is why we have to override from_orm()
            error_trace_as_str = None
            if scan_command_attempt.error_trace:
                error_trace_as_str = ""
                for line in scan_command_attempt.error_trace.format(chain=False):
                    error_trace_as_str += line

            # Create the JSON version of the scan command attempt
            scan_command_attempt_json_cls = field.type_
            all_scan_commands_attempts_json[field_name] = scan_command_attempt_json_cls(
                status=scan_command_attempt.status,
                error_reason=scan_command_attempt.error_reason,
                error_trace=error_trace_as_str,
                result=scan_command_attempt.result,
            )

        return cls(**all_scan_commands_attempts_json)


# Identical fields in the JSON output
_HttpProxySettingsAsJson = pydantic.dataclasses.dataclass(HttpProxySettings, frozen=True)


class _ClientAuthenticationCredentialsAsJson(pydantic.BaseModel):
    # Compared to the ClientAuthenticationCredentials class, this model does not have the key_password field
    certificate_chain_path: Path
    key_path: Path
    key_type: str

    class Config:
        orm_mode = True

    @classmethod
    def from_orm(cls, client_auth_creds: "ClientAuthenticationCredentials") -> "_ClientAuthenticationCredentialsAsJson":
        return cls(
            certificate_chain_path=client_auth_creds.certificate_chain_path,
            key_path=client_auth_creds.key_path,
            key_type=client_auth_creds.key_type.name,
        )


class _ServerTlsProbingResultAsJson(_BaseModelWithOrmModeAndForbid):
    highest_tls_version_supported: str
    cipher_suite_supported: str
    client_auth_requirement: ClientAuthRequirementEnum
    supports_ecdh_key_exchange: bool

    @classmethod
    def from_orm(cls, tls_probing_result: ServerTlsProbingResult) -> "_ServerTlsProbingResultAsJson":
        return cls(
            highest_tls_version_supported=tls_probing_result.highest_tls_version_supported.name,
            cipher_suite_supported=tls_probing_result.cipher_suite_supported,
            client_auth_requirement=tls_probing_result.client_auth_requirement,
            supports_ecdh_key_exchange=tls_probing_result.supports_ecdh_key_exchange,
        )


_ServerTlsProbingResultAsJson.__doc__ = ServerTlsProbingResult.__doc__  # type: ignore


class _ServerNetworkConfigurationAsJson(_BaseModelWithOrmModeAndForbid):
    tls_server_name_indication: str
    tls_opportunistic_encryption: Optional[ProtocolWithOpportunisticTlsEnum] = None
    tls_client_auth_credentials: Optional[_ClientAuthenticationCredentialsAsJson] = None

    xmpp_to_hostname: Optional[str] = None

    network_timeout: int = 5
    network_max_retries: int = 3


_ServerNetworkConfigurationAsJson.__doc__ = ServerNetworkConfiguration.__doc__  # type: ignore


class _ServerNetworkLocationAsJson(_BaseModelWithOrmModeAndForbid):
    hostname: str
    port: int
    connection_type: ConnectionTypeEnum
    ip_address: Optional[str] = None
    http_proxy_settings: Optional[_HttpProxySettingsAsJson] = None  # type: ignore


_ServerNetworkLocationAsJson.__doc__ = ServerNetworkLocation.__doc__  # type: ignore


class ServerScanResultAsJson(_BaseModelWithOrmModeAndForbid):
    uuid: UUID
    server_location: _ServerNetworkLocationAsJson
    network_configuration: _ServerNetworkConfigurationAsJson

    connectivity_status: ServerConnectivityStatusEnum
    connectivity_error_trace: Optional[str]
    connectivity_result: Optional[_ServerTlsProbingResultAsJson]

    scan_status: ServerScanStatusEnum
    scan_result: Optional[AllScanCommandsAttemptsAsJson]

    @classmethod
    def from_orm(cls, server_scan_result: ServerScanResult) -> "ServerScanResultAsJson":
        connectivity_error_trace_as_str = None
        if server_scan_result.connectivity_error_trace:
            connectivity_error_trace_as_str = ""
            for line in server_scan_result.connectivity_error_trace.format(chain=False):
                connectivity_error_trace_as_str += line

        connectivity_result_as_json: Optional[_ServerTlsProbingResultAsJson]
        if server_scan_result.connectivity_result:
            connectivity_result_as_json = _ServerTlsProbingResultAsJson.from_orm(server_scan_result.connectivity_result)
        else:
            connectivity_result_as_json = None

        scan_result_as_json: Optional[AllScanCommandsAttemptsAsJson]
        if server_scan_result.scan_result:
            scan_result_as_json = AllScanCommandsAttemptsAsJson.from_orm(server_scan_result.scan_result)
        else:
            scan_result_as_json = None

        return cls(
            uuid=server_scan_result.uuid,
            server_location=_ServerNetworkLocationAsJson.from_orm(server_scan_result.server_location),
            network_configuration=_ServerNetworkConfigurationAsJson.from_orm(server_scan_result.network_configuration),
            connectivity_status=server_scan_result.connectivity_status,
            connectivity_error_trace=connectivity_error_trace_as_str,
            connectivity_result=connectivity_result_as_json,
            scan_status=server_scan_result.scan_status,
            scan_result=scan_result_as_json,
        )


ServerScanResultAsJson.__doc__ = ServerScanResult.__doc__  # type: ignore


class SslyzeOutputAsJson(pydantic.BaseModel):
    """The "root" dictionary of the JSON output when using the --json command line option.
    """

    server_scan_results: List[ServerScanResultAsJson]
    date_scans_started: datetime
    date_scans_completed: datetime
    sslyze_version: str = __version__
    sslyze_url: str = __url__



import logging
from http.client import HTTPResponse

from dataclasses import dataclass, asdict
from traceback import TracebackException
from urllib.parse import urlsplit

import pydantic
from nassl._nassl import SslError

from sslyze.cli.scan_attempt_json import ScanCommandAttemptAsJson
from sslyze.plugins.plugin_base import (
    ScanCommandImplementation,
    ScanCommandExtraArgument,
    ScanJob,
    ScanCommandResult,
    ScanCommandWrongUsageError,
    ScanCommandCliConnector,
    ScanJobResult,
)
from sslyze.server_connectivity import ServerConnectivityInfo
from sslyze.connection_helpers.http_request_generator import HttpRequestGenerator
from sslyze.connection_helpers.http_response_parser import HttpResponseParser, NotAValidHttpResponseError
from typing import List, Optional


_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ExpectCtHeader:
    """An Expect-CT header parsed from a server's HTTP response.

    Attributes:
        max-age: The content of the max-age field.
        report-uri: The content of report-uri field.
        enforce: True if enforce directive is set.
    """

    max_age: Optional[int]
    report_uri: Optional[str]
    enforce: bool


@dataclass(frozen=True)
class StrictTransportSecurityHeader:
    """A Strict-Transport-Security header parsed from a server's HTTP response.

    Attributes:
        preload: ``True`` if the preload directive is set.
        include_subdomains: ``True`` if the includesubdomains directive is set.
        max_age: The content of the max-age field.
    """

    max_age: Optional[int]
    preload: bool
    include_subdomains: bool


@dataclass(frozen=True)
class HttpHeadersScanResult(ScanCommandResult):
    """The result of testing a server for the presence of security-related HTTP headers.

    Each HTTP header described below will be ``None`` if the server did not return a valid HTTP response, or if the
    server returned an HTTP response without the HTTP header.

    Attributes:
        http_request_sent: The initial HTTP request sent to the server by SSLyze.
        http_error_trace: An error the server returned after receiving the initial HTTP request. If this field is set,
            all the subsequent fields will be ``None`` as SSLyze did not receive a valid HTTP response from the server.
        http_path_redirected_to: The path SSLyze was eventually redirected to after sending the initial HTTP request.
        strict_transport_security_header: The Strict-Transport-Security header returned by the server.
        expect_ct_header: The Expect-CT header returned by the server.
    """

    http_request_sent: str
    http_error_trace: Optional[TracebackException]

    http_path_redirected_to: Optional[str]
    strict_transport_security_header: Optional[StrictTransportSecurityHeader]
    expect_ct_header: Optional[ExpectCtHeader]


class _ExpectCtHeaderAsJson(pydantic.BaseModel):
    max_age: Optional[int]
    report_uri: Optional[str]
    enforce: bool


_ExpectCtHeaderAsJson.__doc__ = ExpectCtHeader.__doc__  # type: ignore


class _StrictTransportSecurityHeaderAsJson(pydantic.BaseModel):
    max_age: Optional[int]
    preload: bool
    include_subdomains: bool


_StrictTransportSecurityHeaderAsJson.__doc__ = StrictTransportSecurityHeader.__doc__  # type: ignore


class HttpHeadersScanResultAsJson(pydantic.BaseModel):
    http_request_sent: str
    http_error_trace: Optional[str]

    http_path_redirected_to: Optional[str]
    strict_transport_security_header: Optional[_StrictTransportSecurityHeaderAsJson]
    expect_ct_header: Optional[_ExpectCtHeaderAsJson]

    class Config:
        orm_mode = True

    @classmethod
    def from_orm(cls, result: HttpHeadersScanResult) -> "HttpHeadersScanResultAsJson":
        http_error_trace_as_str = None
        if result.http_error_trace:
            http_error_trace_as_str = ""
            for line in result.http_error_trace.format(chain=False):
                http_error_trace_as_str += line

        sts_header_json = None
        if result.strict_transport_security_header:
            sts_header_json = _StrictTransportSecurityHeaderAsJson(**asdict(result.strict_transport_security_header))

        ct_header_json = None
        if result.strict_transport_security_header:
            ct_header_json = _ExpectCtHeaderAsJson(**asdict(result.expect_ct_header))

        return cls(
            http_request_sent=result.http_request_sent,
            http_error_trace=http_error_trace_as_str,
            http_path_redirected_to=result.http_path_redirected_to,
            strict_transport_security_header=sts_header_json,
            expect_ct_header=ct_header_json,
        )


HttpHeadersScanResultAsJson.__doc__ = HttpHeadersScanResult.__doc__  # type: ignore


class HttpHeadersScanAttemptAsJson(ScanCommandAttemptAsJson):
    result: Optional[HttpHeadersScanResultAsJson]


class _HttpHeadersCliConnector(ScanCommandCliConnector[HttpHeadersScanResult, None]):

    _cli_option = "http_headers"
    _cli_description = "Test a server for the presence of security-related HTTP headers."

    @classmethod
    def result_to_console_output(cls, result: HttpHeadersScanResult) -> List[str]:
        result_as_txt = [cls._format_title("HTTP Security Headers")]

        # If an error occurred after sending the HTTP request, just display it
        if result.http_error_trace:
            result_as_txt.append(
                cls._format_subtitle("Error: The server did not return a valid HTTP response. Is it an HTTP server?")
            )
            # Extract the last line which contains the reason
            last_line = None
            for line in result.http_error_trace.format(chain=False):
                last_line = line
            if last_line:
                result_as_txt.append(f"     Error details: {last_line.strip()}")

            return result_as_txt

        # HSTS
        result_as_txt.append(cls._format_subtitle("Strict-Transport-Security Header"))
        if not result.strict_transport_security_header:
            result_as_txt.append(cls._format_field("NOT SUPPORTED - Server did not return the header", ""))
        else:
            result_as_txt.append(cls._format_field("Max Age:", str(result.strict_transport_security_header.max_age)))
            result_as_txt.append(
                cls._format_field(
                    "Include Subdomains:", str(result.strict_transport_security_header.include_subdomains)
                )
            )
            result_as_txt.append(cls._format_field("Preload:", str(result.strict_transport_security_header.preload)))

        # Expect-CT
        result_as_txt.extend(["", cls._format_subtitle("Expect-CT Header")])
        if not result.expect_ct_header:
            result_as_txt.append(cls._format_field("NOT SUPPORTED - Server did not return the header", ""))
        else:
            result_as_txt.append(cls._format_field("Max Age:", str(result.expect_ct_header.max_age)))
            result_as_txt.append(cls._format_field("Report- URI:", str(result.expect_ct_header.report_uri)))
            result_as_txt.append(cls._format_field("Enforce:", str(result.expect_ct_header.enforce)))

        return result_as_txt


class HttpHeadersImplementation(ScanCommandImplementation[HttpHeadersScanResult, None]):
    """Test a server for HTTP headers related to security, including HSTS and HPKP.
    """

    cli_connector_cls = _HttpHeadersCliConnector

    @classmethod
    def scan_jobs_for_scan_command(
        cls, server_info: ServerConnectivityInfo, extra_arguments: Optional[ScanCommandExtraArgument] = None
    ) -> List[ScanJob]:
        if extra_arguments:
            raise ScanCommandWrongUsageError("This plugin does not take extra arguments")

        if server_info.network_configuration.tls_opportunistic_encryption:
            raise ScanCommandWrongUsageError("Cannot scan for HTTP headers against a non-HTTP server.")

        return [ScanJob(function_to_call=_retrieve_and_analyze_http_response, function_arguments=[server_info])]

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> HttpHeadersScanResult:
        if len(scan_job_results) != 1:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        return scan_job_results[0].get_result()


def _retrieve_and_analyze_http_response(server_info: ServerConnectivityInfo) -> HttpHeadersScanResult:
    # Send HTTP requests until we no longer received an HTTP redirection, but allow only 4 redirections max
    _logger.info(f"Retrieving HTTP headers from {server_info}")
    redirections_count = 0
    next_location_path: Optional[str] = "/"
    http_error_trace = None

    while next_location_path and redirections_count < 4:
        _logger.info(f"Sending HTTP request to {next_location_path}")
        http_path_redirected_to = next_location_path

        # Perform the TLS handshake
        ssl_connection = server_info.get_preconfigured_tls_connection()
        ssl_connection.connect()

        try:
            # Send an HTTP GET request to the server
            ssl_connection.ssl_client.write(
                HttpRequestGenerator.get_request(
                    host=server_info.network_configuration.tls_server_name_indication, path=next_location_path
                )
            )
            http_response = HttpResponseParser.parse_from_ssl_connection(ssl_connection.ssl_client)

        except (OSError, NotAValidHttpResponseError, SslError) as e:
            # The server closed/rejected the connection, or didn't return a valid HTTP response
            http_error_trace = TracebackException.from_exception(e)

        finally:
            ssl_connection.close()

        if http_error_trace:
            break

        # Handle redirection if there is one
        next_location_path = _detect_http_redirection(
            http_response=http_response,
            server_host_name=server_info.network_configuration.tls_server_name_indication,
            server_port=server_info.server_location.port,
        )
        redirections_count += 1

    # Prepare the results
    initial_http_request = HttpRequestGenerator.get_request(
        host=server_info.network_configuration.tls_server_name_indication, path="/"
    ).decode("ascii")

    if http_error_trace:
        # If the server errored when receiving an HTTP request, return the error as the result
        return HttpHeadersScanResult(
            http_request_sent=initial_http_request,
            http_error_trace=http_error_trace,
            http_path_redirected_to=None,
            strict_transport_security_header=None,
            expect_ct_header=None,
        )
    else:
        # If no HTTP error happened, parse and return each header
        return HttpHeadersScanResult(
            http_request_sent=initial_http_request,
            http_path_redirected_to=http_path_redirected_to,
            http_error_trace=None,
            strict_transport_security_header=_parse_hsts_header_from_http_response(http_response),
            expect_ct_header=_parse_expect_ct_header_from_http_response(http_response),
        )


def _detect_http_redirection(http_response: HTTPResponse, server_host_name: str, server_port: int) -> Optional[str]:
    """If the HTTP response contains a redirection to the same server, return the path to the new location.
    """
    next_location_path = None
    if 300 <= http_response.status < 400:
        location_header = _extract_first_header_value(http_response, "Location")
        if location_header:
            parsed_location = urlsplit(location_header)
            is_relative_url = False if parsed_location.hostname else True
            if is_relative_url:
                # Yes, to a relative URL; follow the redirection
                next_location_path = location_header
            else:
                is_absolute_url_to_same_hostname = parsed_location.hostname == server_host_name
                absolute_url_port = 443 if parsed_location.port is None else parsed_location.port
                is_absolute_url_to_same_port = absolute_url_port == server_port
                if is_absolute_url_to_same_hostname and is_absolute_url_to_same_port:
                    # Yes, to an absolute URL to the same server; follow the redirection
                    next_location_path = f"{parsed_location.path}"
                    if parsed_location.query:
                        next_location_path += f"?{parsed_location.query}"

    return next_location_path


def _extract_first_header_value(response: HTTPResponse, header_name: str) -> Optional[str]:
    raw_header = response.getheader(header_name, None)
    if not raw_header:
        return None

    # Handle headers defined multiple times by picking the first value
    if "," in raw_header:
        raw_header = raw_header.split(",")[0]
    return raw_header


def _parse_hsts_header_from_http_response(response: HTTPResponse) -> Optional[StrictTransportSecurityHeader]:
    raw_hsts_header = _extract_first_header_value(response, "strict-transport-security")
    if not raw_hsts_header:
        return None

    max_age = None
    include_subdomains = False
    preload = False
    for hsts_directive in raw_hsts_header.split(";"):
        hsts_directive = hsts_directive.strip()
        if not hsts_directive:
            # Empty space at the end of the header
            continue

        if "max-age" in hsts_directive:
            max_age = int(hsts_directive.split("max-age=")[1].strip())
        elif "includesubdomains" in hsts_directive.lower():
            # Some websites have a different case for IncludeSubDomains
            include_subdomains = True
        elif "preload" in hsts_directive:
            preload = True
        else:
            _logger.warning(f"Unexpected value in HSTS header: {repr(hsts_directive)}")

    return StrictTransportSecurityHeader(max_age, preload, include_subdomains)


def _parse_expect_ct_header_from_http_response(response: HTTPResponse) -> Optional[ExpectCtHeader]:
    raw_expect_ct_header = _extract_first_header_value(response, "expect-ct")
    if not raw_expect_ct_header:
        return None

    max_age = None
    report_uri = None
    enforce = False
    for expect_ct_directive in raw_expect_ct_header.split(","):
        expect_ct_directive = expect_ct_directive.strip()

        if not expect_ct_directive:
            continue

        if "max-age" in expect_ct_directive:
            max_age = int(expect_ct_directive.split("max-age=")[1].strip())
        elif "report-uri" in expect_ct_directive:
            report_uri = expect_ct_directive.split("report-uri=")[1].strip(' "')
        elif "enforce" in expect_ct_directive:
            enforce = True
        else:
            _logger.warning(f"Unexpected value in Expect-CT header: {repr(expect_ct_directive)}")

    return ExpectCtHeader(max_age, report_uri, enforce)



import socket
from base64 import b64encode
from enum import Enum
from pathlib import Path
from typing import Optional

from urllib.parse import quote

from dataclasses import dataclass
from urllib.parse import urlparse

from nassl.ssl_client import OpenSslFileTypeEnum, SslClient

from sslyze.connection_helpers.opportunistic_tls_helpers import ProtocolWithOpportunisticTlsEnum
from sslyze.errors import InvalidServerNetworkConfigurationError, ServerHostnameCouldNotBeResolved


@dataclass(frozen=True)
class HttpProxySettings:
    hostname: str
    port: int

    basic_auth_user: Optional[str] = None
    basic_auth_password: Optional[str] = None

    @classmethod
    def from_url(cls, proxy_url: str) -> "HttpProxySettings":
        parsed_url = urlparse(proxy_url)
        if not parsed_url.netloc or not parsed_url.hostname:
            raise ValueError("Invalid Proxy URL")

        if parsed_url.scheme == "http":
            default_port = 80
        elif parsed_url.scheme == "https":
            default_port = 443
        else:
            raise ValueError("Invalid URL scheme")

        port = parsed_url.port if parsed_url.port else default_port
        return cls(parsed_url.hostname, port, parsed_url.username, parsed_url.password)

    @property
    def proxy_authorization_header(self) -> Optional[str]:
        if not self.basic_auth_user:
            return None
        if not self.basic_auth_password:
            raise ValueError("No password configured for Basic Auth")

        basic_auth_token = b64encode(f"{quote(self.basic_auth_user)}:{quote(self.basic_auth_password)}".encode("utf-8"))
        return basic_auth_token.decode("utf-8")


class ConnectionTypeEnum(str, Enum):
    DIRECT = "DIRECT"
    VIA_HTTP_PROXY = "VIA_HTTP_PROXY"


@dataclass(frozen=True)
class ServerNetworkLocation:
    """All the information needed to connect to a server.

    Attributes:
        hostname: The server's hostname.
        port: The server's TLS port number.
        connection_type: How sslyze should connect to the server: either directly, or via an HTTP proxy.
        ip_address: The server's IP address; only set if sslyze is connecting directly to the server. If no IP address
            is supplied and connection_type is set to DIRECT, sslyze will automatically lookup one IP address for the
             supplied hostname.
        http_proxy_settings: The HTTP proxy configuration to use in order to tunnel the scans through a proxy; only set
            if sslyze is connecting to the server via an HTTP proxy. The proxy will be responsible for looking up the
            server's IP address and connecting to it.
    """

    hostname: str
    port: int = 443

    # Set if SSLyze is directly connecting to the server ie. connection_type == DIRECT
    ip_address: Optional[str] = None

    # Set if SSLyze is connecting via a proxy ie. connection_type == VIA_HTTP_PROXY
    http_proxy_settings: Optional[HttpProxySettings] = None

    @property
    def connection_type(self) -> ConnectionTypeEnum:
        if self.ip_address:
            return ConnectionTypeEnum.DIRECT
        elif self.http_proxy_settings:
            return ConnectionTypeEnum.VIA_HTTP_PROXY
        else:
            raise ValueError("Should never happen")

    def __post_init__(self) -> None:
        # Official workaround for frozen=True: https://docs.python.org/3/library/dataclasses.html#frozen-instances
        # Store the hostname in ACE format in the case the domain name is unicode
        hostname = self.hostname
        object.__setattr__(self, "hostname", hostname.encode("idna").decode("utf-8"))

        if self.http_proxy_settings and self.ip_address:
            raise ValueError(
                "Cannot supply both ip_address and http_proxy_settings: when using an HTTP proxy, the proxy will lookup"
                "the server's IP address automatically."
            )

        if not self.ip_address and not self.http_proxy_settings:
            # Automatically lookup the IP address
            ip_address = _do_dns_lookup(self.hostname, self.port)
            object.__setattr__(self, "ip_address", ip_address)


def _do_dns_lookup(hostname: str, port: int) -> str:
    try:
        addr_infos = socket.getaddrinfo(hostname, port, socket.AF_UNSPEC, socket.IPPROTO_IP)
    except (socket.gaierror, IndexError, ConnectionError):
        raise ServerHostnameCouldNotBeResolved(f"Could not resolve {hostname}")

    family, socktype, proto, canonname, sockaddr = addr_infos[0]

    # By default use the first DNS entry, IPv4 or IPv6
    tentative_ip_addr = sockaddr[0]

    # But try to use IPv4 if we have both IPv4 and IPv6 addresses, to work around buggy networks
    for family, socktype, proto, canonname, sockaddr in addr_infos:
        if family == socket.AF_INET:
            tentative_ip_addr = sockaddr[0]

    return tentative_ip_addr


@dataclass(frozen=True)
class ClientAuthenticationCredentials:
    """Everything needed by a client to perform SSL/TLS client authentication with the server.

       Attributes:
           certificate_chain_path: Path to the file containing the client's certificate.
           key_path: Path to the file containing the client's private key.
           key_password: The password to decrypt the private key.
           key_type: The format of the key file.
    """

    certificate_chain_path: Path
    key_path: Path
    key_password: str = ""
    key_type: OpenSslFileTypeEnum = OpenSslFileTypeEnum.PEM

    def __post_init__(self) -> None:
        # Try to load the cert and key in OpenSSL; will raise an exception if something is wrong
        SslClient(
            client_certificate_chain=self.certificate_chain_path,
            client_key=self.key_path,
            client_key_type=self.key_type,
            client_key_password=self.key_password,
        )


@dataclass(frozen=True)
class ServerNetworkConfiguration:
    """Additional network settings to provide fine-grained control on how to connect to a specific server.

    Attributes:
        tls_server_name_indication: The hostname to set within the Server Name Indication TLS extension.
        tls_wrapped_protocol: The protocol wrapped in TLS that the server expects. It allows SSLyze to figure out
            how to establish a (Start)TLS connection to the server and what kind of "hello" message
            (SMTP, XMPP, etc.) to send to the server after the handshake was completed. If not supplied, standard
            TLS will be used.
        tls_client_auth_credentials: The client certificate and private key needed to perform mutual authentication
            with the server. If not supplied, SSLyze will attempt to connect to the server without performing
            client authentication.
        xmpp_to_hostname: The hostname to set within the `to` attribute of the XMPP stream. If not supplied, the
            server's hostname will be used. Should only be set if the supplied `tls_wrapped_protocol` is an
            XMPP protocol.
        network_timeout: The timeout (in seconds) to be used when attempting to establish a connection to the
            server.
        network_max_retries: The number of retries SSLyze will perform when attempting to establish a connection
            to the server.
    """

    tls_server_name_indication: str
    tls_opportunistic_encryption: Optional[ProtocolWithOpportunisticTlsEnum] = None
    tls_client_auth_credentials: Optional[ClientAuthenticationCredentials] = None

    xmpp_to_hostname: Optional[str] = None

    network_timeout: int = 5
    network_max_retries: int = 3

    def __post_init__(self) -> None:
        if self.tls_opportunistic_encryption in [
            ProtocolWithOpportunisticTlsEnum.XMPP,
            ProtocolWithOpportunisticTlsEnum.XMPP_SERVER,
        ]:
            if not self.xmpp_to_hostname:
                # Official workaround for frozen: https://docs.python.org/3/library/dataclasses.html#frozen-instances
                # If no XMPP to hostname was supplied, used the ones from SNI
                object.__setattr__(self, "xmpp_to_hostname", self.tls_server_name_indication)
        else:
            if self.xmpp_to_hostname:
                raise InvalidServerNetworkConfigurationError("Can only specify xmpp_to for the XMPP StartTLS protocol.")

    @classmethod
    def default_for_server_location(cls, server_location: ServerNetworkLocation) -> "ServerNetworkConfiguration":
        return cls(tls_server_name_indication=server_location.hostname)



import sys
from os import path, listdir
from pathlib import Path
from typing import List, Tuple, Dict

from setuptools import find_packages

# Setup file based on https://github.com/pypa/sampleproject/blob/master/setup.py
root_path = Path(__file__).parent.absolute()

# For cx_freeze builds, we need a special setup() function
if len(sys.argv) > 1 and sys.argv[1] == "build_exe":
    from cx_Freeze import setup
    from cx_Freeze import Executable
else:
    from setuptools import setup

    # Create fake Executable that does nothing so the setup.py file can be used on Linux
    class Executable:  # type: ignore
        def __init__(self, script, targetName):  # type: ignore
            pass


def get_long_description() -> str:
    path_to_readme = root_path / "README.md"
    return path_to_readme.read_text()


def get_project_info() -> Dict[str, str]:
    project_info: Dict[str, str] = {}
    project_info_path = root_path / "sslyze" / "__version__.py"
    exec(project_info_path.read_text(), project_info)
    return project_info


def get_include_files() -> List[Tuple[str, str]]:
    """"Get the list of trust stores so they properly packaged when doing a cx_freeze build.
    """
    plugin_data_files = []
    trust_stores_pem_path = root_path / "sslyze" / "plugins" / "certificate_info" / "trust_stores" / "pem_files"
    for file in listdir(trust_stores_pem_path):
        file = path.join(trust_stores_pem_path, file)
        if path.isfile(file):  # skip directories
            filename = path.basename(file)
            plugin_data_files.append((file, path.join("pem_files", filename)))
    return plugin_data_files


project_info = get_project_info()


setup(
    name=project_info["__title__"].lower(),
    version=project_info["__version__"],
    description=project_info["__description__"],
    url=project_info["__url__"],
    author=project_info["__author__"],
    author_email=project_info["__author_email__"],
    license=project_info["__license__"],
    python_requires=">=3.7",
    # Pypi metadata
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Natural Language :: French",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: System :: Networking",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: Security",
    ],
    keywords="ssl tls scan security library",
    project_urls={
        "Source": "https://github.com/nabla-c0d3/sslyze",
        "Changelog": "https://github.com/nabla-c0d3/sslyze/releases",
        "Documentation": "https://nabla-c0d3.github.io/sslyze/documentation",
    },
    # Package info
    packages=find_packages(include=["sslyze", "sslyze.*"]),
    package_data={
        "sslyze": ["py.typed"],
        "sslyze.plugins.certificate_info.trust_stores": ["pem_files/*.pem", "pem_files/*.yaml"],
    },
    entry_points={"console_scripts": ["sslyze = sslyze.__main__:main"]},
    # Dependencies
    install_requires=[
        "nassl>=4.0.0,<5.0.0",
        "cryptography>=2.6,<3.5",
        "tls-parser>=1.2.2,<1.3.0",
        "pydantic>=1.7,<1.9",
    ],
    # cx_freeze info for Windows builds with Python embedded
    options={"build_exe": {"packages": ["cffi", "cryptography"], "include_files": get_include_files()}},
    executables=[Executable(path.join("sslyze", "__main__.py"), targetName="sslyze.exe")],
)



import socket
import types
from enum import Enum
from typing import Optional, List, Dict

import binascii
import math
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPublicNumbers
from cryptography.x509 import load_pem_x509_certificate
from nassl._nassl import WantReadError
from nassl.ssl_client import ClientCertificateRequested
from tls_parser.change_cipher_spec_protocol import TlsChangeCipherSpecRecord

from tls_parser.alert_protocol import TlsAlertRecord
from tls_parser.record_protocol import TlsRecordTlsVersionBytes
from tls_parser.exceptions import NotEnoughData
from tls_parser.handshake_protocol import TlsHandshakeRecord, TlsHandshakeTypeByte, TlsRsaClientKeyExchangeRecord
from tls_parser.parser import TlsRecordParser

import tls_parser.tls_version

from sslyze.errors import ServerRejectedTlsHandshake
from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum, ClientAuthRequirementEnum


class RobotScanResultEnum(str, Enum):
    """The result of attempting exploit the ROBOT issue on the server.

    Attributes:
        VULNERABLE_WEAK_ORACLE: The server is vulnerable but the attack would take too long.
        VULNERABLE_STRONG_ORACLE: The server is vulnerable and real attacks are feasible.
        NOT_VULNERABLE_NO_ORACLE: The server supports RSA cipher suites but does not act as an oracle.
        NOT_VULNERABLE_RSA_NOT_SUPPORTED: The server does not supports RSA cipher suites.
        UNKNOWN_INCONSISTENT_RESULTS: Could not determine whether the server is vulnerable or not.
    """

    VULNERABLE_WEAK_ORACLE = "VULNERABLE_WEAK_ORACLE"
    VULNERABLE_STRONG_ORACLE = "VULNERABLE_STRONG_ORACLE"
    NOT_VULNERABLE_NO_ORACLE = "NOT_VULNERABLE_NO_ORACLE"
    NOT_VULNERABLE_RSA_NOT_SUPPORTED = "NOT_VULNERABLE_RSA_NOT_SUPPORTED"
    UNKNOWN_INCONSISTENT_RESULTS = "UNKNOWN_INCONSISTENT_RESULTS"


class RobotPmsPaddingPayloadEnum(Enum):
    VALID = 0
    WRONG_FIRST_TWO_BYTES = 1
    WRONG_POSITION_00 = 2
    NO_00_IN_THE_MIDDLE = 3
    WRONG_VERSION_NUMBER = 4


class _RobotTlsRecordPayloads:

    # From https://github.com/robotattackorg/robot-detect and testssl.sh
    # The high level idea of an oracle attack is to send several payloads that are slightly wrong, in different ways,
    # hoping that the server is going to give a different response (a TLS alert, a connection reset, no data, etc.) for
    # each payload
    _CKE_PAYLOADS_HEX = {
        RobotPmsPaddingPayloadEnum.VALID: "0002{pms_padding}00{tls_version}{pms}",  # noqa: E241
        RobotPmsPaddingPayloadEnum.WRONG_FIRST_TWO_BYTES: "4117{pms_padding}00{tls_version}{pms}",  # noqa: E241
        RobotPmsPaddingPayloadEnum.WRONG_POSITION_00: "0002{pms_padding}11{pms}0011",  # noqa: E241
        RobotPmsPaddingPayloadEnum.NO_00_IN_THE_MIDDLE: "0002{pms_padding}111111{pms}",  # noqa: E241
        RobotPmsPaddingPayloadEnum.WRONG_VERSION_NUMBER: "0002{pms_padding}000202{pms}",  # noqa: E241
    }

    _PMS_HEX = "aa112233445566778899112233445566778899112233445566778899112233445566778899112233445566778899"

    @classmethod
    def get_client_key_exchange_record(
        cls,
        robot_payload_enum: RobotPmsPaddingPayloadEnum,
        tls_version: tls_parser.tls_version.TlsVersionEnum,
        modulus: int,
        exponent: int,
    ) -> TlsRsaClientKeyExchangeRecord:
        """A client key exchange record with a hardcoded pre_master_secret, and a valid or invalid padding.
        """
        pms_padding = cls._compute_pms_padding(modulus)
        tls_version_hex = binascii.b2a_hex(TlsRecordTlsVersionBytes[tls_version.name].value).decode("ascii")

        pms_with_padding_payload = cls._CKE_PAYLOADS_HEX[robot_payload_enum]
        final_pms = pms_with_padding_payload.format(
            pms_padding=pms_padding, tls_version=tls_version_hex, pms=cls._PMS_HEX
        )
        cke_robot_record = TlsRsaClientKeyExchangeRecord.from_parameters(
            tls_version, exponent, modulus, int(final_pms, 16)
        )
        return cke_robot_record

    @staticmethod
    def _compute_pms_padding(modulus: int) -> str:
        # Generate the padding for the pre_master_scecret
        modulus_bit_size = int(math.ceil(math.log(modulus, 2)))
        modulus_byte_size = (modulus_bit_size + 7) // 8
        # pad_len is length in hex chars, so bytelen * 2
        pad_len = (modulus_byte_size - 48 - 3) * 2
        pms_padding_hex = ("abcd" * (pad_len // 2 + 1))[:pad_len]
        return pms_padding_hex

    # Encrypted Finished record corresponding to the PMS below and the ch_def client hello in the ROBOT poc script
    _FINISHED_RECORD = bytearray.fromhex(
        "005091a3b6aaa2b64d126e5583b04c113259c4efa48e40a19b8e5f2542c3b1d30f8d80b7582b72f08b21dfcbff09d4b281676a0fb40"
        "d48c20c4f388617ff5c00808a96fbfe9bb6cc631101a6ba6b6bc696f0"
    )

    @classmethod
    def get_finished_record_bytes(cls, tls_version: tls_parser.tls_version.TlsVersionEnum) -> bytes:
        """The Finished TLS record corresponding to the hardcoded PMS used in the Client Key Exchange record.
        """
        # TODO(AD): The ROBOT poc script uses the same Finished record for all possible client hello (default, GCM,
        # etc.); as the Finished record contains a hashes of all previous records, it will be wrong and will cause
        # servers to send a TLS Alert 20
        # Here just like in the poc script, the Finished message does not match the Client Hello we sent
        return b"\x16" + TlsRecordTlsVersionBytes[tls_version.name].value + cls._FINISHED_RECORD


class RobotServerResponsesAnalyzer:
    def __init__(self, payload_responses: Dict[RobotPmsPaddingPayloadEnum, List[str]], attempts_count: int) -> None:
        # A mapping of a ROBOT payload enum -> a list of two server responses as text
        for server_responses in payload_responses.values():
            if len(server_responses) != attempts_count:
                raise ValueError()
        self._payload_responses = payload_responses
        self._attempts_count = attempts_count

    def compute_result_enum(self) -> RobotScanResultEnum:
        """Look at the server's response to each ROBOT payload and return the conclusion of the analysis.
        """
        # Ensure the results were consistent
        for payload_enum, server_responses in self._payload_responses.items():
            # We ran the check a number of times per payload and the responses should be the same
            if len(set(server_responses)) != 1:
                return RobotScanResultEnum.UNKNOWN_INCONSISTENT_RESULTS

        # Check if the server acts as an oracle by checking if the server replied differently to the payloads
        if len(set([server_responses[0] for server_responses in self._payload_responses.values()])) == 1:
            # All server responses were identical - no oracle
            return RobotScanResultEnum.NOT_VULNERABLE_NO_ORACLE

        # All server responses were NOT identical, server is vulnerable
        # Check to see if it is a weak oracle
        response_1 = self._payload_responses[RobotPmsPaddingPayloadEnum.WRONG_FIRST_TWO_BYTES][0]
        response_2 = self._payload_responses[RobotPmsPaddingPayloadEnum.WRONG_POSITION_00][0]
        response_3 = self._payload_responses[RobotPmsPaddingPayloadEnum.NO_00_IN_THE_MIDDLE][0]

        # From the original script:
        # If the response to the invalid PKCS#1 request (oracle_bad1) is equal to both
        # requests starting with 0002, we have a weak oracle. This is because the only
        # case where we can distinguish valid from invalid requests is when we send
        # correctly formatted PKCS#1 message with 0x00 on a correct position. This
        # makes our oracle weak
        if response_1 == response_2 == response_3:
            return RobotScanResultEnum.VULNERABLE_WEAK_ORACLE
        else:
            return RobotScanResultEnum.VULNERABLE_STRONG_ORACLE


class ServerDoesNotSupportRsa(Exception):
    pass


def test_robot(server_info: ServerConnectivityInfo) -> Dict[RobotPmsPaddingPayloadEnum, str]:
    # Try with TLS 1.2 even if the server supports TLS 1.3 or higher
    if server_info.tls_probing_result.highest_tls_version_supported.value >= TlsVersionEnum.TLS_1_3.value:
        tls_version_to_use = TlsVersionEnum.TLS_1_2
    else:
        tls_version_to_use = server_info.tls_probing_result.highest_tls_version_supported

    rsa_params = None
    # With TLS 1.2 some servers are only vulnerable when using the GCM cipher suites - try them first
    if tls_version_to_use == TlsVersionEnum.TLS_1_2:
        cipher_string = "AES128-GCM-SHA256:AES256-GCM-SHA384"
        rsa_params = _get_rsa_parameters(server_info, tls_version_to_use, cipher_string)

    if rsa_params is None:
        # The attempts with GCM TLS 1.2 RSA cipher suites failed - try the normal RSA cipher suites
        cipher_string = "RSA"
        rsa_params = _get_rsa_parameters(server_info, tls_version_to_use, cipher_string)

    if rsa_params is None:
        # Could not connect to the server using RSA - not vulnerable
        raise ServerDoesNotSupportRsa()

    rsa_modulus = rsa_params.n
    rsa_exponent = rsa_params.e

    # On the first attempt, finish the TLS handshake after sending the Robot payload
    robot_should_complete_handshake = True
    server_responses_per_robot_payloads = _run_oracle_detection(
        server_info, tls_version_to_use, cipher_string, rsa_modulus, rsa_exponent, robot_should_complete_handshake
    )
    return server_responses_per_robot_payloads

    # TODO(AD): The following section was taken from the original ROBOT poc script but makes the scans really slow as it
    # waits for server timeouts
    robot_result_enum = RobotServerResponsesAnalyzer(
        {payload_enum: [response] for payload_enum, response in server_responses_per_robot_payloads.items()}, 1
    ).compute_result_enum()

    if robot_result_enum == RobotScanResultEnum.NOT_VULNERABLE_NO_ORACLE:
        # Try again but this time do not finish the TLS handshake - for some servers it will reveal an oracle
        robot_should_complete_handshake = False
        server_responses_per_robot_payloads = _run_oracle_detection(
            server_info, tls_version_to_use, cipher_string, rsa_modulus, rsa_exponent, robot_should_complete_handshake
        )

    return server_responses_per_robot_payloads


def _run_oracle_detection(
    server_info: ServerConnectivityInfo,
    tls_version_to_use: TlsVersionEnum,
    cipher_string: str,
    rsa_modulus: int,
    rsa_exponent: int,
    should_complete_handshake: bool,
) -> Dict[RobotPmsPaddingPayloadEnum, str]:
    server_responses_per_robot_payloads: Dict[RobotPmsPaddingPayloadEnum, str] = {}
    for payload_enum in RobotPmsPaddingPayloadEnum:
        server_response = _send_robot_payload(
            server_info,
            tls_version_to_use,
            cipher_string,
            payload_enum,
            should_complete_handshake,
            rsa_modulus,
            rsa_exponent,
        )

        server_responses_per_robot_payloads[payload_enum] = server_response
    return server_responses_per_robot_payloads


def _get_rsa_parameters(
    server_info: ServerConnectivityInfo, tls_version: TlsVersionEnum, openssl_cipher_string: str
) -> Optional[RSAPublicNumbers]:
    ssl_connection = server_info.get_preconfigured_tls_connection(
        override_tls_version=tls_version, should_use_legacy_openssl=True,
    )
    ssl_connection.ssl_client.set_cipher_list(openssl_cipher_string)
    parsed_cert = None
    try:
        # Perform the SSL handshake
        ssl_connection.connect()
        cert_as_pem = ssl_connection.ssl_client.get_received_chain()[0]
        parsed_cert = load_pem_x509_certificate(cert_as_pem.encode("ascii"), backend=default_backend())
    except ServerRejectedTlsHandshake:
        # Server does not support RSA cipher suites?
        pass
    except ClientCertificateRequested:
        # AD: The server asked for a client cert. We could still retrieve the server certificate, but it is unclear
        # to me if the ROBOT check is supposed to work even if we do not provide a client cert. My guess is that
        # it should not work since it requires completing a full handshake, which we can't without a client cert.
        # Hence, propagate the error to make the check fail.
        raise
    finally:
        ssl_connection.close()

    if parsed_cert:
        public_key = parsed_cert.public_key()
        if isinstance(public_key, RSAPublicKey):
            return public_key.public_numbers()
        else:
            return None
    else:
        return None


def _send_robot_payload(
    server_info: ServerConnectivityInfo,
    tls_version_to_use: TlsVersionEnum,
    rsa_cipher_string: str,
    robot_payload_enum: RobotPmsPaddingPayloadEnum,
    robot_should_finish_handshake: bool,
    rsa_modulus: int,
    rsa_exponent: int,
) -> str:
    # Do a handshake which each record and keep track of what the server returned
    ssl_connection = server_info.get_preconfigured_tls_connection(override_tls_version=tls_version_to_use)

    # Replace nassl.sslClient.do_handshake() with a ROBOT checking SSL handshake so that all the SSLyze
    # options (startTLS, proxy, etc.) still work
    ssl_connection.ssl_client.do_handshake = types.MethodType(  # type: ignore
        do_handshake_with_robot, ssl_connection.ssl_client
    )
    ssl_connection.ssl_client.set_cipher_list(rsa_cipher_string)

    # Compute the  payload
    tls_parser_tls_version: tls_parser.tls_version.TlsVersionEnum
    if tls_version_to_use == TlsVersionEnum.SSL_3_0:
        tls_parser_tls_version = tls_parser.tls_version.TlsVersionEnum.SSLV3
    elif tls_version_to_use == TlsVersionEnum.TLS_1_0:
        tls_parser_tls_version = tls_parser.tls_version.TlsVersionEnum.TLSV1
    elif tls_version_to_use == TlsVersionEnum.TLS_1_1:
        tls_parser_tls_version = tls_parser.tls_version.TlsVersionEnum.TLSV1_1
    elif tls_version_to_use == TlsVersionEnum.TLS_1_2:
        tls_parser_tls_version = tls_parser.tls_version.TlsVersionEnum.TLSV1_2
    else:
        raise ValueError("Should never happen")

    cke_payload = _RobotTlsRecordPayloads.get_client_key_exchange_record(
        robot_payload_enum, tls_parser_tls_version, rsa_modulus, rsa_exponent
    )

    # H4ck: we need to pass some arguments to the handshake but there is no simple way to do it; we use an attribute
    ssl_connection.ssl_client._robot_cke_record = cke_payload  # type: ignore
    ssl_connection.ssl_client._robot_should_finish_handshake = robot_should_finish_handshake  # type: ignore

    server_response = ""
    try:
        # Start the SSL handshake
        ssl_connection.connect()
    except ServerResponseToRobot as e:
        # Should always be thrown
        server_response = e.server_response
    except socket.timeout:
        # https://github.com/nabla-c0d3/sslyze/issues/361
        server_response = "Connection timed out"
    except ServerRejectedTlsHandshake:
        if server_info.tls_probing_result.client_auth_requirement != ClientAuthRequirementEnum.DISABLED:
            # This error happens when scanning an nginx server with client authentication required;
            # If the server asks for a client cert, we cannot check for ROBOT as the check needs to complete full
            # handshakes. https://github.com/nabla-c0d3/sslyze/issues/484
            raise ClientCertificateRequested(ca_list=[])
        else:
            raise
    finally:
        ssl_connection.close()

    return server_response


class ServerResponseToRobot(Exception):
    def __init__(self, server_response: str) -> None:
        # Could be a TLS alert or some data, always as text so we can easily detect different responses
        self.server_response = server_response


def do_handshake_with_robot(self):  # type: ignore
    """Modified do_handshake() to send a ROBOT payload and return the result.
    """
    try:
        # Start the handshake using nassl - will throw WantReadError right away
        self._ssl.do_handshake()
    except WantReadError:
        # Send the Client Hello
        len_to_read = self._network_bio.pending()
        while len_to_read:
            # Get the data from the SSL engine
            handshake_data_out = self._network_bio.read(len_to_read)
            # Send it to the peer
            self._sock.send(handshake_data_out)
            len_to_read = self._network_bio.pending()

    # Retrieve the server's response - directly read the underlying network socket
    # Retrieve data until we get to the ServerHelloDone
    # The server may send back a ServerHello, an Alert or a CertificateRequest first
    did_receive_hello_done = False
    remaining_bytes = b""
    while not did_receive_hello_done:
        try:
            tls_record, len_consumed = TlsRecordParser.parse_bytes(remaining_bytes)
            remaining_bytes = remaining_bytes[len_consumed::]
        except NotEnoughData:
            # Try to get more data
            raw_ssl_bytes = self._sock.recv(16381)
            if not raw_ssl_bytes:
                # No data?
                break

            remaining_bytes = remaining_bytes + raw_ssl_bytes
            continue

        if isinstance(tls_record, TlsHandshakeRecord):
            # Does the record contain a ServerDone message?
            for handshake_message in tls_record.subprotocol_messages:
                if handshake_message.handshake_type == TlsHandshakeTypeByte.SERVER_DONE:
                    did_receive_hello_done = True
                    break
            # If not, it could be a ServerHello, Certificate or a CertificateRequest if the server requires client auth
        elif isinstance(tls_record, TlsAlertRecord):
            # Server returned a TLS alert
            break
        else:
            raise ValueError("Unknown record? Type {}".format(tls_record.header.type))

    if did_receive_hello_done:
        # Send a special Client Key Exchange Record as the payload
        self._sock.send(self._robot_cke_record.to_bytes())

        if self._robot_should_finish_handshake:
            # Then send a CCS record
            ccs_record = TlsChangeCipherSpecRecord.from_parameters(
                tls_version=tls_parser.tls_version.TlsVersionEnum[self._ssl_version.name]
            )
            self._sock.send(ccs_record.to_bytes())

            # Lastly send a Finished record
            finished_record_bytes = _RobotTlsRecordPayloads.get_finished_record_bytes(self._ssl_version)
            self._sock.send(finished_record_bytes)

        # Return whatever the server sent back by raising an exception
        # The goal is to detect similar/different responses
        while True:
            try:
                tls_record, len_consumed = TlsRecordParser.parse_bytes(remaining_bytes)
                remaining_bytes = remaining_bytes[len_consumed::]
            except NotEnoughData:
                # Try to get more data
                try:
                    raw_ssl_bytes = self._sock.recv(16381)
                    if not raw_ssl_bytes:
                        # No data?
                        raise ServerResponseToRobot("No data")
                except socket.error as e:
                    # Server closed the connection after receiving the CCS payload
                    raise ServerResponseToRobot("socket.error {}".format(str(e)))

                remaining_bytes = remaining_bytes + raw_ssl_bytes
                continue

            if isinstance(tls_record, TlsAlertRecord):
                raise ServerResponseToRobot(
                    "TLS Alert {} {}".format(tls_record.alert_description, tls_record.alert_severity)
                )
            else:
                break

        raise ServerResponseToRobot("Ok")



from abc import ABC, abstractmethod

from sslyze import ServerScanRequest, ServerTlsProbingResult, ServerScanResult
from sslyze.errors import ConnectionToServerFailed


# Could be turned into a typing.Protocol once we stop supporting Python 3.7
class ScannerObserver(ABC):
    @abstractmethod
    def server_connectivity_test_error(
        self, server_scan_request: ServerScanRequest, connectivity_error: ConnectionToServerFailed
    ) -> None:
        """The Scanner found a server that it could not connect to; no scans will be performed against this server.
        """

    @abstractmethod
    def server_connectivity_test_completed(
        self, server_scan_request: ServerScanRequest, connectivity_result: ServerTlsProbingResult
    ) -> None:
        """The Scanner found a server that it was able to connect to; scans will be run against this server.
        """

    @abstractmethod
    def server_scan_completed(self, server_scan_result: ServerScanResult) -> None:
        """The Scanner finished scanning one single server.
        """

    @abstractmethod
    def all_server_scans_completed(self) -> None:
        """The Scanner finished scanning all the supplied servers and will now exit.
        """
